"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initServingValidatorsInterval = exports.clearServingValidatorsInterval = exports.provideGlobalAccountReportRequest = exports.provideAccountDataByListRequest = exports.provideAccountDataRequest = exports.validateGlobalAccountReportRequest = exports.validateAccountDataByListRequest = exports.validateAccountDataRequest = exports.servingValidators = void 0;
const Crypto = require("../Crypto");
const Account = require("../dbstore/accounts");
const Logger = require("../Logger");
const Config_1 = require("../Config");
const Utils = require("../Utils");
const GlobalAccount_1 = require("../GlobalAccount");
const NodeList = require("../NodeList");
const Cycles_1 = require("./Cycles");
// This has to align with the queue sit time in the validator
const QUEUE_SIT_TIME = 6 * 1000; // 6 seconds
// maxValidatorsToServe
exports.servingValidators = new Map(); // key: validatorKey, value: lastServedTimestamp
const SERVING_VALIDATOR_TIMEOUT = 10 * 1000; // 10 seconds
let servingValidatorsRemovalInterval;
const validateAccountDataRequest = (payload) => {
    if (Cycles_1.currentNetworkMode !== 'restore') {
        return { success: false, error: 'Account data can only be requested in restore mode!' };
    }
    let err = Utils.validateTypes(payload, {
        accountStart: 's',
        accountEnd: 's',
        tsStart: 'n',
        maxRecords: 'n',
        offset: 'n',
        accountOffset: 's',
        sign: 'o',
    });
    if (err) {
        return { success: false, error: err };
    }
    const { accountStart, accountEnd, tsStart, maxRecords, offset, accountOffset, sign } = payload;
    err = Utils.validateTypes(sign, { owner: 's', sig: 's' });
    if (err) {
        return { success: false, error: 'Invalid sign object attached' };
    }
    const nodePublicKey = sign.owner;
    if (!NodeList.byPublicKey.has(nodePublicKey)) {
        return { success: false, error: 'This node is not found in the nodelist!' };
    }
    if (!exports.servingValidators.has(nodePublicKey) && exports.servingValidators.size >= Config_1.config.maxValidatorsToServe) {
        return {
            success: false,
            error: 'Archiver is busy serving other validators at the moment!',
        };
    }
    if (accountStart.length !== 64 || accountEnd.length !== 64 || accountStart > accountEnd) {
        return { success: false, error: 'Invalid account range' };
    }
    if (Number.isNaN(tsStart) || tsStart < 0 || tsStart > Date.now()) {
        return { success: false, error: 'Invalid start timestamp' };
    }
    if (Number.isNaN(maxRecords) || maxRecords < 1) {
        return { success: false, error: 'Invalid max records' };
    }
    if (Number.isNaN(offset) || offset < 0) {
        return { success: false, error: 'Invalid offset' };
    }
    if (accountOffset && accountOffset.length !== 64) {
        return { success: false, error: 'Invalid account offset' };
    }
    if (!Crypto.verify(payload)) {
        return { success: false, error: 'Invalid signature' };
    }
    exports.servingValidators.set(nodePublicKey, Date.now());
    return { success: true };
};
exports.validateAccountDataRequest = validateAccountDataRequest;
const validateAccountDataByListRequest = (payload) => {
    if (Cycles_1.currentNetworkMode !== 'restore') {
        return { success: false, error: 'Account data by list can only be requested in restore mode!' };
    }
    let err = Utils.validateTypes(payload, {
        accountIds: 'a',
        sign: 'o',
    });
    if (err) {
        return { success: false, error: err };
    }
    const { accountIds, sign } = payload;
    err = Utils.validateTypes(sign, { owner: 's', sig: 's' });
    if (err) {
        return { success: false, error: 'Invalid sign object attached' };
    }
    const nodePublicKey = sign.owner;
    if (!NodeList.byPublicKey.has(nodePublicKey)) {
        return { success: false, error: 'This node is not found in the nodelist!' };
    }
    // TODO: Add max limit check for accountIds list query
    if (accountIds.length === 0 || accountIds.some((accountId) => accountId.length !== 64)) {
        return { success: false, error: 'Invalid account ids' };
    }
    if (!Crypto.verify(payload)) {
        return { success: false, error: 'Invalid signature' };
    }
    return { success: true };
};
exports.validateAccountDataByListRequest = validateAccountDataByListRequest;
const validateGlobalAccountReportRequest = (payload) => {
    let err = Utils.validateTypes(payload, {
        sign: 'o',
    });
    if (err) {
        return { success: false, error: err };
    }
    const { sign } = payload;
    err = Utils.validateTypes(sign, { owner: 's', sig: 's' });
    if (err) {
        return { success: false, error: 'Invalid sign object attached' };
    }
    if (!Crypto.verify(payload)) {
        return { success: false, error: 'Invalid signature' };
    }
    return { success: true };
};
exports.validateGlobalAccountReportRequest = validateGlobalAccountReportRequest;
/**
 *
 * This function is contructed to provide data in similar way as the `getAccountDataByRangeSmart` function in the validator
 * @param payload
 * @returns GetAccountDataByRangeSmart
 */
const provideAccountDataRequest = async (payload) => {
    const wrappedAccounts = [];
    const wrappedAccounts2 = []; // We might not need to provide data to this
    let lastUpdateNeeded = false;
    let highestTs = 0;
    let delta = 0;
    const { accountStart, accountEnd, tsStart, maxRecords, offset, accountOffset } = payload;
    const tsEnd = Date.now();
    // Query from Shardeum->queryAccountsEntryByRanges3 fn
    // const query = `SELECT * FROM accountsEntry
    // WHERE (timestamp, accountId) >= (${tsStart}, "${accountOffset}")
    // AND timestamp < ${tsEnd}
    // AND accountId <= "${accountEnd}" AND accountId >= "${accountStart}"
    // ORDER BY timestamp, accountId  LIMIT ${maxRecords}`
    const safeSkip = Number.isInteger(offset) ? offset : 0;
    const safeLimit = Number.isInteger(maxRecords) ? maxRecords : 100;
    const sqlPrefix = `SELECT * FROM accounts WHERE `;
    const queryString = `accountId BETWEEN ? AND ? AND timestamp BETWEEN ? AND ? ORDER BY timestamp ASC, accountId ASC LIMIT ${safeLimit}`;
    const offsetCondition = ` OFFSET ${safeSkip}`;
    let sql = sqlPrefix;
    let values = [];
    if (accountOffset) {
        sql += `accountId >= ? AND `;
        values.push(accountOffset);
    }
    sql += queryString;
    values.push(accountStart, accountEnd, tsStart, tsEnd);
    if (!accountOffset)
        sql += offsetCondition;
    let accounts = await Account.fetchAccountsBySqlQuery(sql, values);
    for (const account of accounts) {
        wrappedAccounts.push({
            accountId: account.accountId,
            stateId: account.hash,
            data: account.data,
            timestamp: account.timestamp,
        });
    }
    if (wrappedAccounts.length === 0) {
        lastUpdateNeeded = true;
    }
    else {
        // see if our newest record is new enough
        highestTs = 0;
        for (const account of wrappedAccounts) {
            if (account.timestamp > highestTs) {
                highestTs = account.timestamp;
            }
        }
        delta = tsEnd - highestTs;
        // Logger.mainLogger.debug('Account Data received', StringUtils.safeStringify(payload))
        // Logger.mainLogger.debug(
        //   'delta ' + delta,
        //   'tsEnd ' + tsEnd,
        //   'highestTs ' + highestTs,
        //   delta < QUEUE_SIT_TIME * 2
        // )
        if (delta < QUEUE_SIT_TIME * 2) {
            const tsStart2 = highestTs;
            const tsEnd2 = Date.now();
            sql = sqlPrefix + queryString + offsetCondition;
            values = [accountStart, accountEnd, tsStart2, tsEnd2];
            accounts = await Account.fetchAccountsBySqlQuery(sql, values);
            for (const account of accounts) {
                wrappedAccounts2.push({
                    accountId: account.accountId,
                    stateId: account.hash,
                    data: account.data,
                    timestamp: account.timestamp,
                });
            }
            lastUpdateNeeded = true;
        }
    }
    return { wrappedAccounts, lastUpdateNeeded, wrappedAccounts2, highestTs, delta };
};
exports.provideAccountDataRequest = provideAccountDataRequest;
const provideAccountDataByListRequest = async (payload) => {
    const { accountIds } = payload;
    const wrappedAccounts = [];
    // todo: does this even work right now?
    const sql = `SELECT * FROM accounts WHERE accountId IN (?)`;
    const accounts = await Account.fetchAccountsBySqlQuery(sql, accountIds);
    for (const account of accounts) {
        wrappedAccounts.push({
            accountId: account.accountId,
            stateId: account.hash,
            data: account.data,
            timestamp: account.timestamp,
        });
    }
    return wrappedAccounts;
};
exports.provideAccountDataByListRequest = provideAccountDataByListRequest;
const provideGlobalAccountReportRequest = async () => {
    const result = { ready: true, combinedHash: '', accounts: [] };
    for (const [key, value] of GlobalAccount_1.globalAccountsMap.entries()) {
        result.accounts.push({ id: key, hash: value.hash, timestamp: value.timestamp });
    }
    result.accounts.sort(Utils.byIdAsc);
    result.combinedHash = Crypto.hashObj(result);
    return result;
};
exports.provideGlobalAccountReportRequest = provideGlobalAccountReportRequest;
// Remove validators from the list that have not requested data over 10 seconds, that way we can serve new validators
const clearTimeoutServingValidators = () => {
    const now = Date.now();
    for (const [validatorKey, lastServedTimestamp] of exports.servingValidators.entries()) {
        if (now - lastServedTimestamp > SERVING_VALIDATOR_TIMEOUT) {
            Logger.mainLogger.debug('Removing validator from serving list', validatorKey);
            exports.servingValidators.delete(validatorKey);
        }
    }
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Serving validators', exports.servingValidators.size, exports.servingValidators);
};
const clearServingValidatorsInterval = () => {
    Logger.mainLogger.debug('clearServingValidatorsInterval');
    if (servingValidatorsRemovalInterval) {
        clearInterval(servingValidatorsRemovalInterval);
        servingValidatorsRemovalInterval = null;
    }
};
exports.clearServingValidatorsInterval = clearServingValidatorsInterval;
const initServingValidatorsInterval = () => {
    Logger.mainLogger.debug('initServingValidatorsInterval');
    if (!servingValidatorsRemovalInterval)
        servingValidatorsRemovalInterval = setInterval(clearTimeoutServingValidators, SERVING_VALIDATOR_TIMEOUT);
};
exports.initServingValidatorsInterval = initServingValidatorsInterval;
//# sourceMappingURL=AccountDataProvider.js.map