"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.verifyMultiSigs = exports.createDirectories = exports.byIdAsc = exports.attempt = exports.isUndefined = exports.validateTypes = exports.sleep = exports.getRandomItemFromArr = exports.binarySearch = exports.computeMedian = exports.insertSorted = exports.deepCopy = exports.sequentialQuery = exports.robustQuery = exports.robustPromiseAll = exports.shuffleArray = exports.Comparison = exports.getRandom = void 0;
const fs = require("fs");
const path = require("path");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const util = require("util");
const Logger = require("./Logger");
const ethers_1 = require("ethers");
// From: https://stackoverflow.com/a/19270021
function getRandom(arr, n) {
    let len = arr.length;
    const taken = new Array(len);
    if (n > len) {
        n = len;
    }
    const result = new Array(n);
    while (n--) {
        const x = Math.floor(Math.random() * len);
        // eslint-disable-next-line security/detect-object-injection
        result[n] = arr[x in taken ? taken[x] : x];
        // eslint-disable-next-line security/detect-object-injection
        taken[x] = --len in taken ? taken[len] : len;
    }
    return result;
}
exports.getRandom = getRandom;
var Comparison;
(function (Comparison) {
    Comparison[Comparison["BETTER"] = 0] = "BETTER";
    Comparison[Comparison["EQUAL"] = 1] = "EQUAL";
    Comparison[Comparison["WORSE"] = 2] = "WORSE";
    Comparison[Comparison["ABORT"] = 3] = "ABORT";
})(Comparison = exports.Comparison || (exports.Comparison = {}));
function shuffleArray(array) {
    for (let i = array.length - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1));
        [array[i], array[j]] = [array[j], array[i]];
    }
}
exports.shuffleArray = shuffleArray;
const robustPromiseAll = async (promises) => {
    // This is how we wrap a promise to prevent it from rejecting directing in the Promise.all and causing a short circuit
    const wrapPromise = async (promise) => {
        // We are trying to await the promise, and catching any rejections
        // We return an array, the first index being resolve, and the second being an error
        try {
            const result = await promise;
            return [result];
        }
        catch (e) {
            return [null, e];
        }
    };
    const wrappedPromises = [];
    // We wrap all the promises we received and push them to an array to be Promise.all'd
    for (const promise of promises) {
        wrappedPromises.push(wrapPromise(promise));
    }
    const resolved = [];
    const errors = [];
    // We await the wrapped promises to finish resolving or rejecting
    const wrappedResults = await Promise.all(wrappedPromises);
    // We iterate over all the results, checking if they resolved or rejected
    for (const wrapped of wrappedResults) {
        const [result, err] = wrapped;
        // If there was an error, we push it to our errors array
        if (err) {
            errors.push(err);
            continue;
        }
        // Otherwise, we were able to resolve so we push it to the resolved array
        resolved.push(result);
    }
    // We return two arrays, one of the resolved promises, and one of the errors
    return [resolved, errors];
};
exports.robustPromiseAll = robustPromiseAll;
class Tally {
    constructor(winCount, equalFn) {
        this.winCount = winCount;
        this.equalFn = equalFn;
        this.items = [];
    }
    add(newItem, node) {
        if (newItem === null)
            return null;
        // We search to see if we've already seen this item before
        for (const item of this.items) {
            // If the value of the new item is not equal to the current item, we continue searching
            if (!this.equalFn(newItem, item.value))
                continue;
            // If the new item is equal to the current item in the list,
            // we increment the current item's counter and add the current node to the list
            item.count++;
            item.nodes.push(node);
            // Here we check our win condition if the current item's counter was incremented
            // If we meet the win requirement, we return an array with the value of the item,
            // and the list of nodes who voted for that item
            if (item.count >= this.winCount) {
                return item;
            }
            // Otherwise, if the win condition hasn't been met,
            // We return null to indicate no winner yet
            return null;
        }
        // If we made it through the entire items list without finding a match,
        // We create a new item and set the count to 1
        const newTallyItem = { value: newItem, count: 1, nodes: [node] };
        this.items.push(newTallyItem);
        // Finally, we check to see if the winCount is 1,
        // and return the item we just created if that is the case
        if (this.winCount === 1)
            return newTallyItem;
        else
            return null;
    }
    getHighestCount() {
        if (!this.items.length)
            return 0;
        let highestCount = 0;
        for (const item of this.items) {
            if (item.count > highestCount) {
                highestCount = item.count;
            }
        }
        return highestCount;
    }
    getHighestCountItem() {
        if (!this.items.length)
            return null;
        let highestCount = 0;
        let highestIndex = 0;
        let i = 0;
        for (const item of this.items) {
            if (item.count > highestCount) {
                highestCount = item.count;
                highestIndex = i;
            }
            i += 1;
        }
        // eslint-disable-next-line security/detect-object-injection
        return this.items[highestIndex];
    }
}
async function robustQuery(nodes = [], queryFn, equalityFn = util.isDeepStrictEqual, redundancy = 3, shuffleNodes = false, // Set to true if nodes need to be shuffled before querying
delayTimeInMS = 0, // Add a delay after we have queried half of the nodes
disableFailLog = false) {
    if (nodes.length === 0)
        throw new Error('No nodes given.');
    if (typeof queryFn !== 'function') {
        throw new Error(`Provided queryFn ${queryFn} is not a valid function.`);
    }
    if (redundancy < 1)
        redundancy = 3;
    if (redundancy > nodes.length)
        redundancy = nodes.length;
    const responses = new Tally(redundancy, equalityFn);
    let errors = 0;
    // [TODO] - Change the way we shuffle the array.
    //     This is not scaleable, if the size of the nodes array is over 100 we should create an array of
    //     indexes and shuffle that. Or maybe use a function that treats the array as a ring and starts at
    //     a random offset in the ring and a random direction. Or use a function that visits every element
    //     in the array once in a random order.
    nodes = [...nodes];
    if (shuffleNodes === true) {
        shuffleArray(nodes);
    }
    const nodeCount = nodes.length;
    const queryNodes = async (nodes) => {
        // Wrap the query so that we know which node it's coming from
        const wrappedQuery = async (node) => {
            const response = await queryFn(node);
            return { response, node };
        };
        // We create a promise for each of the first `redundancy` nodes in the shuffled array
        const queries = [];
        for (let i = 0; i < nodes.length; i++) {
            // eslint-disable-next-line security/detect-object-injection
            const node = nodes[i];
            queries.push(wrappedQuery(node));
        }
        const [results, errs] = await (0, exports.robustPromiseAll)(queries);
        let finalResult;
        for (const result of results) {
            const { response, node } = result;
            if (responses === null)
                continue; // ignore null response; can be null if we tried to query ourself
            finalResult = responses.add(response, node);
            if (finalResult)
                break;
        }
        for (const err of errs) {
            Logger.mainLogger.error('p2p/Utils:robustQuery:queryNodes:', err);
            errors += 1;
        }
        if (!finalResult)
            return null;
        return finalResult;
    };
    let finalResult = null;
    let tries = 0;
    while (!finalResult) {
        tries += 1;
        const toQuery = redundancy - responses.getHighestCount();
        if (nodes.length < toQuery) {
            if (!disableFailLog)
                Logger.mainLogger.error('In robustQuery stopping since we ran out of nodes to query.');
            break;
        }
        if (delayTimeInMS > 0 && Math.ceil(nodeCount / 2) >= nodes.length) {
            await sleep(delayTimeInMS);
        }
        const nodesToQuery = nodes.splice(0, toQuery);
        finalResult = await queryNodes(nodesToQuery);
        if (tries >= 20) {
            Logger.mainLogger.error('In robustQuery stopping after 20 tries.');
            console.trace();
            break;
        }
    }
    if (finalResult) {
        // Logger.mainLogger.debug(`In robustQuery stopping since we got a finalResult:${StringUtils.safeStringify(finalResult)}`)
        return finalResult;
    }
    else {
        // TODO:  We return the item that had the most nodes reporting it. However, the caller should know
        //        what the count was. We should return [item, count] so that caller gets both.
        //        This change would require also changing all the places it is called.
        if (!disableFailLog)
            Logger.mainLogger.error(`Could not get ${redundancy} ${redundancy > 1 ? 'redundant responses' : 'response'} from ${nodeCount} ${nodeCount !== 1 ? 'nodes' : 'node'}. Encountered ${errors} query errors.`);
        console.trace();
        return responses.getHighestCountItem();
    }
}
exports.robustQuery = robustQuery;
async function sequentialQuery(nodes, queryFn, verifyFn = () => true) {
    nodes = [...nodes];
    shuffleArray(nodes);
    let result;
    const errors = [];
    for (const node of nodes) {
        try {
            const response = await queryFn(node);
            if (verifyFn(response) === false) {
                errors.push({
                    node,
                    error: new Error('Response failed verifyFn'),
                    response,
                });
                continue;
            }
            result = response;
        }
        catch (error) {
            errors.push({
                node,
                error: error,
            });
        }
    }
    return {
        result,
        errors,
    };
}
exports.sequentialQuery = sequentialQuery;
const deepCopy = (obj) => {
    if (typeof obj !== 'object') {
        throw Error('Given element is not of type object.');
    }
    return lib_types_1.Utils.safeJsonParse(lib_types_1.Utils.safeStringify(obj));
};
exports.deepCopy = deepCopy;
const insertSorted = function (arr, item, comparator) {
    let i = (0, exports.binarySearch)(arr, item, comparator);
    if (i < 0) {
        i = -1 - i;
    }
    arr.splice(i, 0, item);
};
exports.insertSorted = insertSorted;
const computeMedian = (arr = [], sort = true) => {
    if (sort) {
        arr.sort((a, b) => a - b);
    }
    const len = arr.length;
    switch (len) {
        case 0: {
            return 0;
        }
        case 1: {
            return arr[0];
        }
        default: {
            const mid = len / 2;
            if (len % 2 === 0) {
                // eslint-disable-next-line security/detect-object-injection
                return arr[mid];
            }
            else {
                return (arr[Math.floor(mid)] + arr[Math.ceil(mid)]) / 2;
            }
        }
    }
};
exports.computeMedian = computeMedian;
const binarySearch = function (arr, el, comparator) {
    if (comparator == null) {
        // Emulate the default Array.sort() comparator
        comparator = (a, b) => {
            return a.toString() > b.toString() ? 1 : a.toString() < b.toString() ? -1 : 0;
        };
    }
    let m = 0;
    let n = arr.length - 1;
    while (m <= n) {
        const k = (n + m) >> 1;
        // eslint-disable-next-line security/detect-object-injection
        const cmp = comparator(el, arr[k]);
        if (cmp > 0) {
            m = k + 1;
        }
        else if (cmp < 0) {
            n = k - 1;
        }
        else {
            return k;
        }
    }
    return -m - 1;
};
exports.binarySearch = binarySearch;
// fail safe and fast
// this function will pick non-repeating multiple random elements from an array if given amount n > 1
//(partial) fisher-yates shuffle
function getRandomItemFromArr(arr, nodeRejectPercentage = 0, n = 1) {
    if (!Array.isArray(arr))
        return undefined;
    if (arr.length === 0)
        return undefined;
    let result = new Array(n), len = arr.length;
    const taken = new Array(len);
    const oldNodesToAvoid = Math.floor((nodeRejectPercentage * len) / 100);
    if (n <= 1) {
        const randomIndex = Math.floor(Math.random() * arr.length);
        // eslint-disable-next-line security/detect-object-injection
        return [arr[randomIndex]];
        // we can throw an error but no
        // let's just return one random item in this case for the safety
    }
    while (n-- && len > 0) {
        const x = Math.floor(oldNodesToAvoid + Math.random() * (len - oldNodesToAvoid));
        // eslint-disable-next-line security/detect-object-injection
        result[n] = arr[x in taken ? taken[x] : x];
        // eslint-disable-next-line security/detect-object-injection
        taken[x] = --len in taken ? taken[len] : len;
    }
    // remove empty slots
    if (len === 0) {
        result = result.filter((item) => item !== undefined);
    }
    return result;
}
exports.getRandomItemFromArr = getRandomItemFromArr;
async function sleep(time) {
    Logger.mainLogger.debug('sleeping for', time);
    return new Promise((resolve) => {
        setTimeout(() => {
            resolve(true);
        }, time);
    });
}
exports.sleep = sleep;
/*
inp is the input object to be checked
def is an object defining the expected input
{name1:type1, name1:type2, ...}
name is the name of the field
type is a string with the first letter of 'string', 'number', 'Bigint', 'boolean', 'array' or 'object'
type can end with '?' to indicate that the field is optional and not required
---
Example of def:
{fullname:'s', age:'s?',phone:'sn'}
---
Returns a string with the first error encountered or and empty string ''.
Errors are: "[name] is required" or "[name] must be, [type]"
*/
function validateTypes(inp, def = {}) {
    if (inp === undefined)
        return 'input is undefined';
    if (inp === null)
        return 'input is null';
    if (typeof inp !== 'object')
        return 'input must be object, not ' + typeof inp;
    const map = {
        string: 's',
        number: 'n',
        boolean: 'b',
        bigint: 'B',
        array: 'a',
        object: 'o',
    };
    const imap = {
        s: 'string',
        n: 'number',
        b: 'boolean',
        B: 'bigint',
        a: 'array',
        o: 'object',
    };
    const fields = Object.keys(def);
    for (const name of fields) {
        // eslint-disable-next-line security/detect-object-injection
        const types = def[name];
        const opt = types.substr(-1, 1) === '?' ? 1 : 0;
        // eslint-disable-next-line security/detect-object-injection
        if (inp[name] === undefined && !opt)
            return name + ' is required';
        // eslint-disable-next-line security/detect-object-injection
        if (inp[name] !== undefined) {
            // eslint-disable-next-line security/detect-object-injection
            if (inp[name] === null && !opt)
                return name + ' cannot be null';
            let found = 0;
            let be = '';
            for (let t = 0; t < types.length - opt; t++) {
                // eslint-disable-next-line security/detect-object-injection
                let it = map[typeof inp[name]];
                // eslint-disable-next-line security/detect-object-injection
                it = Array.isArray(inp[name]) ? 'a' : it;
                const is = types.substr(t, 1);
                if (it === is) {
                    found = 1;
                    break;
                    // eslint-disable-next-line security/detect-object-injection
                }
                else
                    be += ', ' + imap[is];
            }
            if (!found)
                return name + ' must be' + be;
        }
    }
    return '';
}
exports.validateTypes = validateTypes;
/**
 * Checks whether the given thing is undefined
 */
function isUndefined(thing) {
    return typeof thing === 'undefined';
}
exports.isUndefined = isUndefined;
/**
 * Attempts to execute a given asynchronous function up to a certain number of retries upon failure.
 *
 * @template T The type of the resolved value of the input function.
 * @param {() => Promise<T>} fn - The asynchronous function to execute. This function should return a Promise that resolves to a value of type `T`.
 * @param {AttemptOptions} options - Optional. Options passed to change the behavior of this function. See the `AttemptOptions` interface in this same file for details.
 * @returns {Promise<T>} A Promise that resolves to the return value of the input function, if successful.
 * @throws Will throw an error if the function fails all attempts. The error will be the last error thrown by the input function.
 *
 * Code is duplicated from shardus-global-server.
 */
async function attempt(fn, options) {
    // fallback to option defaults if needed
    const maxRetries = options?.maxRetries || 3;
    const delay = options?.delay || 2000;
    const logPrefix = options?.logPrefix || 'attempt';
    // initialize our lastError variable
    let lastError = new Error('out of retries');
    // loop until we're successful
    for (let i = 0; i < maxRetries; i++) {
        try {
            // run the function and return the result. if the funciton fails,
            // we'll catch it below
            return await fn();
        }
        catch (e) {
            // log the error
            console.error(`${logPrefix}: attempt failure #${i + 1}: ${e.message}`);
            // save the error in case we need to throw it later
            lastError = e;
            // sleep before trying again
            await sleep(delay);
            continue;
        }
    }
    // log that we've run out of attempts
    console.error(`${logPrefix}: giving up`);
    // think fast!
    throw lastError;
}
exports.attempt = attempt;
const byIdAsc = (a, b) => a.id === b.id ? 0 : a.id < b.id ? -1 : 1;
exports.byIdAsc = byIdAsc;
function createDirectories(pathname) {
    const __dirname = path.resolve();
    pathname = pathname.replace(/^\.*\/|\/?[^/]+\.[a-z]+|\/$/g, ''); // Remove leading directory markers, and remove ending /file-name.extension
    fs.mkdirSync(path.resolve(__dirname, pathname), { recursive: true }); // eslint-disable-line security/detect-non-literal-fs-filename
}
exports.createDirectories = createDirectories;
function verifyMultiSigs(rawPayload, sigs, allowedPubkeys, minSigRequired, requiredSecurityLevel) {
    if (!rawPayload || !sigs || !allowedPubkeys || !Array.isArray(sigs)) {
        return { isValid: false, validCount: 0 };
    }
    if (sigs.length < minSigRequired)
        return { isValid: false, validCount: 0 };
    // no reason to allow more signatures than allowedPubkeys exist
    // this also prevent loop exhaustion
    if (sigs.length > Object.keys(allowedPubkeys).length)
        return { isValid: false, validCount: 0 };
    let validSigs = 0;
    const payload_hash = ethers_1.ethers.keccak256(ethers_1.ethers.toUtf8Bytes(lib_types_1.Utils.safeStringify(rawPayload)));
    const seen = new Set();
    for (let i = 0; i < sigs.length; i++) {
        /* eslint-disable security/detect-object-injection */
        // The sig owner has not been seen before
        // The sig owner is listed on the server
        // The sig owner has enough security clearance
        // The signature is valid
        // The signature is not a duplicate
        if (!seen.has(sigs[i].owner.toLowerCase()) &&
            allowedPubkeys[sigs[i].owner] &&
            allowedPubkeys[sigs[i].owner] >= requiredSecurityLevel &&
            ethers_1.ethers.verifyMessage(payload_hash, sigs[i].sig).toLowerCase() === sigs[i].owner.toLowerCase()) {
            validSigs++;
            seen.add(sigs[i].owner.toLowerCase());
        }
        /* eslint-enable security/detect-object-injection */
        if (validSigs >= minSigRequired)
            break;
    }
    return {
        isValid: validSigs >= minSigRequired,
        validCount: validSigs
    };
}
exports.verifyMultiSigs = verifyMultiSigs;
//# sourceMappingURL=Utils.js.map