"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.addHashesGossip = exports.convertStateMetadataToHashArray = exports.sendGossip = void 0;
const Crypto = require("../Crypto");
const StateMetaData = require("./StateMetaData");
const State = require("../State");
const P2P = require("../P2P");
const Config_1 = require("../Config");
const Logger = require("../Logger");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const gossipCollector = new Map();
async function sendGossip(type, payload) {
    let archivers = [...State.activeArchivers];
    if (archivers.length === 0)
        return;
    const gossipPayload = {
        type,
        data: payload,
        sender: Config_1.config.ARCHIVER_PUBLIC_KEY,
    };
    archivers = archivers.sort((a, b) => Number(a.publicKey) - Number(b.publicKey));
    // TODO: check if need to select random archivers instead of sending to all other archivers
    const recipients = archivers.filter((a) => a.publicKey !== Config_1.config.ARCHIVER_PUBLIC_KEY);
    if (recipients.length === 0) {
        Logger.mainLogger.debug('There is no other archivers to send our gossip');
        return;
    }
    try {
        Logger.mainLogger.debug(`GossipingIn ${type} request to these nodes: ${lib_types_1.Utils.safeStringify(recipients.map((node) => node.ip + ':' + node.port + `/gossip-${type}`))}`);
        await tell(recipients, `gossip-${type}`, gossipPayload);
    }
    catch (ex) {
        Logger.mainLogger.debug(ex);
        Logger.mainLogger.debug('Fail to gossip');
    }
}
exports.sendGossip = sendGossip;
async function tell(nodes, route, message) {
    let InternalTellCounter = 0;
    const promises = [];
    for (const node of nodes) {
        InternalTellCounter++;
        if (Config_1.config.VERBOSE) {
            Logger.mainLogger.debug(`InternalTellCounter: ${InternalTellCounter}`);
        }
        const url = `http://${node.ip}:${node.port}/${route}`;
        try {
            const promise = P2P.postJson(url, message);
            promise.catch((err) => {
                Logger.mainLogger.error(`Unable to tell node ${node.ip}: ${node.port}`, err);
            });
            promises.push(promise);
        }
        catch (e) {
            Logger.mainLogger.error('Error', e);
        }
    }
    try {
        await Promise.all(promises);
    }
    catch (err) {
        Logger.mainLogger.error('Network: ' + err);
    }
}
function convertStateMetadataToHashArray(STATE_METATDATA) {
    const hashCollector = {};
    STATE_METATDATA.stateHashes.forEach((h) => {
        if (!hashCollector[h.counter]) {
            hashCollector[h.counter] = {
                counter: h.counter,
            };
        }
        hashCollector[h.counter]['stateHashes'] = h;
    });
    STATE_METATDATA.receiptHashes.forEach((h) => {
        if (!hashCollector[h.counter]) {
            hashCollector[h.counter] = {
                counter: h.counter,
            };
        }
        hashCollector[h.counter]['receiptHashes'] = h;
    });
    STATE_METATDATA.summaryHashes.forEach((h) => {
        if (!hashCollector[h.counter]) {
            hashCollector[h.counter] = {
                counter: h.counter,
            };
        }
        hashCollector[h.counter]['summaryHashes'] = h;
    });
    return Object.values(hashCollector);
}
exports.convertStateMetadataToHashArray = convertStateMetadataToHashArray;
function addHashesGossip(sender, gossip) {
    const counter = gossip.counter;
    if (gossipCollector.has(counter)) {
        const existingGossips = gossipCollector.get(counter);
        // eslint-disable-next-line security/detect-object-injection
        existingGossips[sender] = gossip;
    }
    else {
        const obj = {};
        // eslint-disable-next-line security/detect-object-injection
        obj[sender] = gossip;
        gossipCollector.set(counter, obj);
    }
    const totalGossip = gossipCollector.get(counter);
    if (totalGossip && Object.keys(totalGossip).length > 0.5 * State.activeArchivers.length) {
        setTimeout(() => {
            processGossip(counter);
            gossipCollector.delete(counter);
        }, 500);
    }
}
exports.addHashesGossip = addHashesGossip;
function processGossip(counter) {
    Logger.mainLogger.debug('Processing gossips for counter', counter, gossipCollector.get(counter));
    const gossips = gossipCollector.get(counter);
    if (!gossips) {
        return;
    }
    const ourHashes = StateMetaData.StateMetaDataMap.get(counter);
    const gossipCounter = {};
    for (const sender in gossips) {
        /* eslint-disable security/detect-object-injection */
        const hashedGossip = Crypto.hashObj(gossips[sender]);
        if (!gossipCounter[hashedGossip]) {
            gossipCounter[hashedGossip] = {
                count: 1,
                gossip: gossips[sender],
            };
            // To count our StateMetaData also
            if (hashedGossip === Crypto.hashObj(ourHashes)) {
                gossipCounter[hashedGossip].count += 1;
            }
        }
        else {
            gossipCounter[hashedGossip].count += 1;
            /* eslint-enable security/detect-object-injection */
        }
    }
    const gossipWithHighestCount = [];
    let highestCount = 0;
    let hashWithHighestCounter;
    for (const key in gossipCounter) {
        // eslint-disable-next-line security/detect-object-injection
        if (gossipCounter[key].count > highestCount) {
            // eslint-disable-next-line security/detect-object-injection
            gossipWithHighestCount.push(gossipCounter[key].gossip);
            hashWithHighestCounter = key;
            // eslint-disable-next-line security/detect-object-injection
            highestCount = gossipCounter[key].count;
        }
    }
    if (!ourHashes) {
        Logger.mainLogger.error(`Unable to find our stored statemetadata hashes for counter ${counter}`);
        return;
    }
    if (hashWithHighestCounter && hashWithHighestCounter !== Crypto.hashObj(ourHashes)) {
        if (gossipWithHighestCount.length === 0) {
            return;
        }
        Logger.mainLogger.error('our hash is different from other archivers hashes. Storing the correct hashes');
        Logger.mainLogger.debug('gossipWithHighestCount', gossipWithHighestCount[0].summaryHashes);
        StateMetaData.processStateMetaData({ gossipWithHighestCount });
        StateMetaData.replaceDataSender(StateMetaData.currentDataSender);
    }
}
//# sourceMappingURL=Gossip.js.map