"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.statisticsInstance = void 0;
const path = require("path");
const fs = require("fs");
const stream_1 = require("stream");
const events_1 = require("events");
const utils = require("../Utils");
const nestedCounters_1 = require("../profiler/nestedCounters");
class Statistics extends events_1.EventEmitter {
    constructor(baseDir, config, { counters = [], watchers = {}, timers = [], manualStats = [], }, context) {
        super();
        this.intervalDuration = config.interval || 1;
        this.intervalDuration = this.intervalDuration * 1000;
        this.context = context;
        this.counterDefs = counters;
        this.watcherDefs = watchers;
        this.timerDefs = timers;
        this.manualStatDefs = manualStats;
        this.initialize();
        this.interval = null;
        this.snapshotWriteFns = [];
        this.stream = null;
        this.streamIsPushable = false;
        // Assigning the instance of the Statistics class to the global variable 'statisticsInstance'.
        // This is done to implement the singleton pattern, where only one instance of the Statistics class is created and used throughout the application.
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        exports.statisticsInstance = this;
        if (config.save) {
            // Pipe stream to file
            const file = path.join(baseDir, 'statistics.tsv');
            // Ignoring the non-literal fs filename rule because 'baseDir' is a trusted source and not user input.
            // eslint-disable-next-line security/detect-non-literal-fs-filename
            const fileWriteStream = fs.createWriteStream(file);
            const statsReadStream = this.getStream();
            statsReadStream.pipe(fileWriteStream);
        }
    }
    initialize() {
        this.counters = this._initializeCounters(this.counterDefs);
        this.watchers = this._initializeWatchers(this.watcherDefs, this.context);
        this.timers = this._initializeTimers(this.timerDefs);
        this.manualStats = this._initializeManualStats(this.manualStatDefs);
    }
    getStream() {
        this.stream = new stream_1.Readable();
        this.stream._read = () => {
            this.streamIsPushable = true;
        };
        return this.stream;
    }
    writeOnSnapshot(writeFn, context) {
        this.snapshotWriteFns.push(writeFn.bind(context));
    }
    startSnapshots() {
        console.log('Starting statistics snapshots...');
        const tabSeperatedHeaders = 'Name\tValue\tTime\n';
        this._pushToStream(tabSeperatedHeaders);
        if (!this.interval)
            this.interval = setInterval(this._takeSnapshot.bind(this), this.intervalDuration);
    }
    stopSnapshots() {
        if (this.interval)
            clearInterval(this.interval);
        this.interval = null;
    }
    // Increments the given CounterRing's count
    incrementCounter(counterName) {
        // eslint-disable-next-line security/detect-object-injection
        const counter = this.counters[counterName];
        if (!counter)
            throw new Error(`Counter '${counterName}' is undefined.`);
        counter.increment();
        nestedCounters_1.nestedCountersInstance.countEvent('statistics', counterName);
    }
    setManualStat(manualStatName, value) {
        // eslint-disable-next-line security/detect-object-injection
        const ring = this.manualStats[manualStatName];
        if (!ring)
            throw new Error(`manualStat '${manualStatName}' is undefined.`);
        ring.manualSetValue(value);
        //nestedCountersInstance.countEvent('statistics', manualStatName)
    }
    // Returns the current count of the given CounterRing
    getCurrentCount(counterName) {
        // eslint-disable-next-line security/detect-object-injection
        const counter = this.counters[counterName];
        if (!counter)
            throw new Error(`Counter '${counterName}' is undefined.`);
        return counter.count;
    }
    // Returns the current total of the given CounterRing
    getCounterTotal(counterName) {
        // eslint-disable-next-line security/detect-object-injection
        const counter = this.counters[counterName];
        if (!counter)
            throw new Error(`Counter '${counterName}' is undefined.`);
        return counter.total;
    }
    // Returns the result of the given WatcherRings watchFn
    getWatcherValue(watcherName) {
        // eslint-disable-next-line security/detect-object-injection
        const watcher = this.watchers[watcherName];
        if (!watcher)
            throw new Error(`Watcher '${watcherName}' is undefined.`);
        return watcher.watchFn();
    }
    // Starts an entry for the given id in the given TimerRing
    startTimer(timerName, id) {
        // eslint-disable-next-line security/detect-object-injection
        const timer = this.timers[timerName];
        if (!timer)
            throw new Error(`Timer '${timerName}' is undefined.`);
        timer.start(id);
    }
    // Stops an entry for the given id in the given TimerRing
    stopTimer(timerName, id) {
        // eslint-disable-next-line security/detect-object-injection
        const timer = this.timers[timerName];
        if (!timer)
            throw new Error(`Timer '${timerName}' is undefined.`);
        timer.stop(id);
    }
    // Returns the current average of all elements in the given WatcherRing, CounterRing, or TimerRing
    getAverage(name) {
        const ringHolder = this.counters[name] || this.watchers[name] || this.timers[name] || this.manualStats[name]; // eslint-disable-line security/detect-object-injection
        if (!ringHolder.ring)
            throw new Error(`Ring holder '${name}' is undefined.`);
        return ringHolder.ring.average();
    }
    getMultiStatReport(name) {
        const ringHolder = this.counters[name] || this.watchers[name] || this.timers[name] || this.manualStats[name]; // eslint-disable-line security/detect-object-injection
        if (!ringHolder.ring)
            throw new Error(`Ring holder '${name}' is undefined.`);
        return ringHolder.ring.multiStats();
    }
    // Returns the value of the last element of the given WatcherRing, CounterRing, or TimerRing
    getPreviousElement(name) {
        const ringHolder = this.counters[name] || this.watchers[name] || this.timers[name]; // eslint-disable-line security/detect-object-injection
        if (!ringHolder.ring)
            throw new Error(`Ring holder '${name}' is undefined.`);
        return ringHolder.ring.previous();
    }
    _initializeCounters(counterDefs = []) {
        const counters = {};
        for (const name of counterDefs) {
            // eslint-disable-next-line security/detect-object-injection
            counters[name] = new CounterRing(60);
        }
        return counters;
    }
    _initializeWatchers(watcherDefs = {}, context) {
        const watchers = {};
        for (const name in watcherDefs) {
            // eslint-disable-next-line security/detect-object-injection
            const watchFn = watcherDefs[name];
            // eslint-disable-next-line security/detect-object-injection
            watchers[name] = new WatcherRing(60, watchFn, context);
        }
        return watchers;
    }
    _initializeTimers(timerDefs = []) {
        const timers = {};
        for (const name of timerDefs) {
            // eslint-disable-next-line security/detect-object-injection
            timers[name] = new TimerRing(60);
        }
        return timers;
    }
    _initializeManualStats(counterDefs = []) {
        const manualStats = {};
        for (const name of counterDefs) {
            // eslint-disable-next-line security/detect-object-injection
            manualStats[name] = new ManualRing(60); //should it be a config
        }
        return manualStats;
    }
    _takeSnapshot() {
        const time = new Date().toISOString();
        let tabSeperatedValues = '';
        for (const counter in this.counters) {
            // eslint-disable-next-line security/detect-object-injection
            this.counters[counter].snapshot();
            tabSeperatedValues += `${counter}-average\t${this.getAverage(counter)}\t${time}\n`;
            tabSeperatedValues += `${counter}-total\t${this.getCounterTotal(counter)}\t${time}\n`;
        }
        for (const watcher in this.watchers) {
            // eslint-disable-next-line security/detect-object-injection
            this.watchers[watcher].snapshot();
            tabSeperatedValues += `${watcher}-average\t${this.getAverage(watcher)}\t${time}\n`;
            tabSeperatedValues += `${watcher}-value\t${this.getWatcherValue(watcher)}\t${time}\n`;
        }
        for (const timer in this.timers) {
            // eslint-disable-next-line security/detect-object-injection
            this.timers[timer].snapshot();
            tabSeperatedValues += `${timer}-average\t${this.getAverage(timer) / 1000}\t${time}\n`;
        }
        for (const writeFn of this.snapshotWriteFns) {
            tabSeperatedValues += writeFn();
        }
        this._pushToStream(tabSeperatedValues);
        this.emit('snapshot');
    }
    _pushToStream(data) {
        if (this.stream && this.streamIsPushable) {
            this.streamIsPushable = this.stream.push(data);
        }
    }
}
class Ring {
    constructor(length) {
        this.elements = new Array(length);
        this.index = 0;
    }
    multiStats() {
        let sum = 0;
        let total = 0;
        let min = Number.MAX_VALUE;
        let max = Number.MIN_VALUE;
        const allVals = [];
        for (const element of this.elements) {
            if (_exists(element)) {
                const val = Number(element);
                sum += val;
                total++;
                if (val < min) {
                    min = val;
                }
                if (val > max) {
                    max = val;
                }
                allVals.push(val);
            }
        }
        const avg = total > 0 ? sum / total : 0;
        return { min, max, avg, allVals };
    }
    save(value) {
        this.elements[this.index] = value;
        this.index = ++this.index % this.elements.length;
    }
    average() {
        let sum = 0;
        let total = 0;
        console.log('elements', this.elements);
        for (const element of this.elements) {
            if (_exists(element)) {
                sum += Number(element);
                total++;
            }
        }
        return total > 0 ? sum / total : 0;
    }
    previous() {
        const prevIndex = (this.index < 1 ? this.elements.length : this.index) - 1;
        // eslint-disable-next-line security/detect-object-injection
        return this.elements[prevIndex] || 0;
    }
}
class CounterRing {
    constructor(length) {
        this.count = 0;
        this.total = 0;
        this.ring = new Ring(length);
    }
    increment() {
        ++this.count;
        ++this.total;
    }
    snapshot() {
        this.ring.save(this.count);
        this.count = 0;
    }
}
class WatcherRing {
    constructor(length, watchFn, context) {
        this.watchFn = watchFn.bind(context);
        this.ring = new Ring(length);
    }
    snapshot() {
        const value = this.watchFn();
        this.ring.save(value);
    }
}
class TimerRing {
    constructor(length) {
        this.ids = {};
        this.ring = new Ring(length);
    }
    start(id) {
        // eslint-disable-next-line security/detect-object-injection
        if (!this.ids[id]) {
            // eslint-disable-next-line security/detect-object-injection
            this.ids[id] = Date.now();
        }
    }
    stop(id) {
        // eslint-disable-next-line security/detect-object-injection
        const entry = this.ids[id];
        if (entry) {
            // eslint-disable-next-line security/detect-object-injection
            delete this.ids[id];
        }
    }
    snapshot() {
        // Calc median duration of all entries in ids
        const durations = [];
        for (const id in this.ids) {
            // eslint-disable-next-line security/detect-object-injection
            const startTime = this.ids[id];
            // console.log('START_TIME ', startTime, 'ID', id)
            const duration = Date.now() - startTime;
            utils.insertSorted(durations, duration, (a, b) => a - b);
        }
        const median = utils.computeMedian(durations, false);
        // Save median
        this.ring.save(median);
    }
}
class ManualRing {
    constructor(length) {
        this.ring = new Ring(length);
    }
    manualSetValue(value) {
        this.ring.save(value);
    }
    snapshot() {
        // intentionally left blank
    }
}
/**
 * Check for a variable that is not undefined or null
 * @param thing The parameter to check
 */
function _exists(thing) {
    return typeof thing !== 'undefined' && thing !== null;
}
exports.default = Statistics;
//# sourceMappingURL=index.js.map