"use strict";
/**
 * SyncV2 a p2p module that contains all of the functionality for the new
 * Node List Sync v2.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncTxList = exports.syncV2 = void 0;
const neverthrow_1 = require("neverthrow");
const queries_1 = require("./queries");
const State_1 = require("../State");
const NodeList_1 = require("../NodeList");
const NodeList = require("../NodeList");
const verify_1 = require("./verify");
const Logger = require("../Logger");
const ServiceQueue = require("../ServiceQueue");
/**
 * Given a list of archivers, queries each one until one returns an active node list.
 *
 * The endpoint queried does not return a *full* list of nodes. It's a partial
 * list that will be enough to use in robust queries.
 */
async function getActiveListFromSomeArchiver(archivers) {
    for (const archiver of archivers) {
        try {
            const nodeList = await (0, NodeList_1.getActiveNodeListFromArchiver)(archiver);
            if (nodeList) {
                return nodeList;
            }
        }
        catch (e) {
            console.warn(`failed to get active node list from archiver ${archiver.ip}:${archiver.port}: ${e}`);
            continue;
        }
    }
    // all archivers have failed at this point
    throw new Error('no archiver could return an active node list');
}
/**
 * Synchronizes the NodeList and gets the latest CycleRecord from other validators.
 */
function syncV2(activeArchivers) {
    return neverthrow_1.ResultAsync.fromPromise(getActiveListFromSomeArchiver(activeArchivers), (e) => e).andThen((nodeList) => syncValidatorList(nodeList).andThen(([validatorList, validatorListHash]) => syncArchiverList(nodeList).andThen(([archiverList, archiverListHash]) => syncStandbyNodeList(nodeList).andThen(([standbyList, standbyListHash]) => syncTxList(nodeList).andThen((txList) => syncLatestCycleRecordAndMarker(nodeList).andThen(([cycle, cycleMarker]) => {
        Logger.mainLogger.debug('syncV2: validatorList', validatorList);
        // additional checks to make sure the list hashes in the cycle
        // matches the hash for the validator list retrieved earlier
        if (cycle.nodeListHash !== validatorListHash) {
            return (0, neverthrow_1.errAsync)(new Error(`validator list hash from received cycle (${cycle.nodeListHash}) does not match the hash received from robust query (${validatorListHash})`));
        }
        if (cycle.standbyNodeListHash !== standbyListHash) {
            return (0, neverthrow_1.errAsync)(new Error(`standby list hash from received cycle (${cycle.nodeListHash}) does not match the hash received from robust query (${validatorListHash})`));
        }
        if (cycle.archiverListHash !== archiverListHash) {
            return (0, neverthrow_1.errAsync)(new Error(`archiver list hash from received cycle (${cycle.archiverListHash}) does not match the hash received from robust query (${archiverListHash})`));
        }
        // validatorList and standbyList need to be transformed into a ConsensusNodeInfo[]
        const syncingNodeList = [];
        const activeNodeList = [];
        for (const node of validatorList) {
            if (node.status === 'selected' || node.status === 'syncing' || node.status === 'ready') {
                syncingNodeList.push({
                    publicKey: node.publicKey,
                    ip: node.externalIp,
                    port: node.externalPort,
                    id: node.id,
                });
            }
            else if (node.status === 'active') {
                activeNodeList.push({
                    publicKey: node.publicKey,
                    ip: node.externalIp,
                    port: node.externalPort,
                    id: node.id,
                });
            }
        }
        const standbyNodeList = standbyList.map((joinRequest) => ({
            publicKey: joinRequest.nodeInfo.publicKey,
            ip: joinRequest.nodeInfo.externalIp,
            port: joinRequest.nodeInfo.externalPort,
        }));
        NodeList.addNodes(NodeList.NodeStatus.SYNCING, syncingNodeList);
        NodeList.addNodes(NodeList.NodeStatus.ACTIVE, activeNodeList);
        NodeList.addStandbyNodes(standbyNodeList);
        // add txList
        ServiceQueue.setTxList(txList);
        // reset the active archivers list with the new list
        (0, State_1.resetActiveArchivers)(archiverList);
        // return a cycle that we'll store in the database
        return (0, neverthrow_1.okAsync)({
            ...cycle,
            marker: cycleMarker,
        });
    }))))));
}
exports.syncV2 = syncV2;
/**
 * This function synchronizes a validator list from `activeNodes`.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest node list hash.
 * After obtaining the hash, it retrieves the full node list from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.NodeListTypes.Node[], Error>} - A ResultAsync object. On success, it will contain
 * an array of Node objects, and on error, it will contain an Error object. The function is asynchronous
 * and can be awaited.
 */
function syncValidatorList(activeNodes) {
    // run a robust query for the lastest node list hash
    return (0, queries_1.robustQueryForValidatorListHash)(activeNodes).andThen(({ value, winningNodes }) => 
    // get full node list from one of the winning nodes
    (0, queries_1.getValidatorListFromNode)(winningNodes[0], value.nodeListHash).andThen((validatorList) => (0, verify_1.verifyValidatorList)(validatorList, value.nodeListHash).map(() => [validatorList, value.nodeListHash])));
}
/**
 * This function synchronizes a standby node list from `activeNodes`.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest node list hash.
 * After obtaining the hash, it retrieves the full node list from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.NodeListTypes.Node[], Error>} - A ResultAsync object. On success, it will contain
 * an array of Node objects, and on error, it will contain an Error object. The function is asynchronous
 * and can be awaited.
 */
function syncStandbyNodeList(activeNodes) {
    // run a robust query for the lastest archiver list hash
    return (0, queries_1.robustQueryForStandbyNodeListHash)(activeNodes).andThen(({ value, winningNodes }) => 
    // get full standby list from one of the winning nodes
    (0, queries_1.getStandbyNodeListFromNode)(winningNodes[0], value.standbyNodeListHash).andThen((standbyList) => (0, neverthrow_1.okAsync)([standbyList, value.standbyNodeListHash])));
}
function syncTxList(activeNodes) {
    return (0, queries_1.robustQueryForTxListHash)(activeNodes).andThen(({ value, winningNodes }) => (0, queries_1.getTxListFromNode)(winningNodes[0], value.txListHash).andThen((txList) => (0, verify_1.verifyTxList)(txList, value.txListHash).map(() => txList)));
}
exports.syncTxList = syncTxList;
/**
 * Synchronizes the latest cycle record from a list of active nodes.
 *
 * @param {P2PTypes.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest cycle record hash.
 * After obtaining the hash, it retrieves the current cycle data from one of the winning nodes.
 *
 * @returns {ResultAsync<P2PTypes.CycleCreatorTypes.CycleRecord, Error>} - A ResultAsync object.
 * On success, it will contain a CycleRecord object, and on error, it will contain an Error object.
 * The function is asynchronous and can be awaited.
 */
function syncLatestCycleRecordAndMarker(activeNodes) {
    // run a robust query for the latest cycle record hash
    return (0, queries_1.robustQueryForCycleRecordHash)(activeNodes).andThen(({ value, winningNodes }) => 
    // get current cycle record from node
    (0, queries_1.getCurrentCycleDataFromNode)(winningNodes[0], value.currentCycleHash).andThen((cycleRecord) => (0, verify_1.verifyCycleRecord)(cycleRecord, value.currentCycleHash).map(() => [cycleRecord, value.currentCycleHash])));
}
/**
 * This function queries for an archiver list from other active nodes.
 *
 * @param {P2P.SyncTypes.ActiveNode[]} activeNodes - An array of active nodes to be queried.
 * The function first performs a robust query for the latest archiver list hash.
 * Then, it requests a full list from one of the winning nodes using the hash
 * retrieved. The node receiving the request may or may not have the list whose
 * hash matches the one requested.
 *
 * @returns {ResultAsync<[P2P.ArchiversTypes.JoinedArchiver[], hexstring], Error>} - A ResultAsync object. On success, it will contain an array of
 * JoinedArchiver objects and the archiver list hash, and on error, it will contain an Error object. The function is asynchronous and can be awaited.
 */
function syncArchiverList(activeNodes) {
    // run a robust query for the lastest archiver list hash
    return (0, queries_1.robustQueryForArchiverListHash)(activeNodes).andThen(({ value, winningNodes }) => 
    // get full archiver list from one of the winning nodes
    (0, queries_1.getArchiverListFromNode)(winningNodes[0], value.archiverListHash).andThen((archiverList) => 
    // verify a hash of the retrieved archiver list matches the hash from before.
    // if it does, return the archiver list
    (0, verify_1.verifyArchiverList)(archiverList, value.archiverListHash).map(() => [archiverList, value.archiverListHash])));
}
//# sourceMappingURL=index.js.map