"use strict";
/**
 * `verify` submodule. Defines functions used to verify objects against
 * given hashes.
 *
 * This module is functionally identical to the one in shardus-global-server/p2p/SyncV2/verify.ts.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.verifyCycleRecord = exports.verifyTxList = exports.verifyArchiverList = exports.verifyStandbyList = exports.verifyValidatorList = void 0;
const lib_crypto_utils_1 = require("@shardeum-foundation/lib-crypto-utils");
const neverthrow_1 = require("neverthrow");
const Cycles_1 = require("../Data/Cycles");
/**
 * Verifies if the hash of a given object matches the expected hash.
 *
 * This function hashes a 'HashableObject' and compares it to an expected hash value. If the hashes match, it returns a successful result with value true. If the hashes don't match, it returns an error with a detailed message describing the mismatch.
 *
 * @param object - The object to be hashed and verified.
 * @param expectedHash - The expected hash string to compare with the hash of the object.
 * @param [objectName='some object'] - An optional name for the object, used in the error message in case of a hash mismatch.
 *
 * @returns Returns a Result object. On successful hash verification, returns 'ok' with value true. On mismatch, returns 'err' with an Error object detailing the mismatch.
 */
function verify(object, expectedHash, objectName = 'some object') {
    const newHash = (0, lib_crypto_utils_1.hashObj)(object);
    return newHash === expectedHash
        ? (0, neverthrow_1.ok)(true)
        : (0, neverthrow_1.err)(new Error(`hash mismatch for ${objectName}: expected ${expectedHash}, got ${newHash}`));
}
/** Verifies that the hash of the validator list matches the expected hash. */
function verifyValidatorList(validatorList, expectedHash) {
    return verify(validatorList, expectedHash, 'validator list');
}
exports.verifyValidatorList = verifyValidatorList;
/** Verifies that the hash of the standby list matches the expected hash. */
function verifyStandbyList(standbyList, expectedHash) {
    return verify(standbyList, expectedHash, 'standby list');
}
exports.verifyStandbyList = verifyStandbyList;
/** Verifies that the hash of the archiver list matches the expected hash. */
function verifyArchiverList(archiverList, expectedHash) {
    return verify(archiverList, expectedHash, 'archiver list');
}
exports.verifyArchiverList = verifyArchiverList;
/** Verifies that the hash of the tx list matches the expected hash. */
function verifyTxList(txList, expectedHash) {
    const actualHash = (0, lib_crypto_utils_1.hashObj)(txList);
    // verify that the hash of the CycleRecord matches the expected hash
    if (actualHash !== expectedHash)
        return (0, neverthrow_1.err)(new Error(`hash mismatch for txList: expected ${expectedHash}, got ${actualHash}`));
    return (0, neverthrow_1.ok)(true);
}
exports.verifyTxList = verifyTxList;
/** Verifies that the hash of the cycle record matches the expected hash. */
function verifyCycleRecord(cycleRecord, expectedHash) {
    const actualHash = (0, Cycles_1.computeCycleMarker)(cycleRecord);
    // verify that the hash of the CycleRecord matches the expected hash
    if (actualHash !== expectedHash)
        return (0, neverthrow_1.err)(new Error(`hash mismatch for cycle: expected ${expectedHash}, got ${actualHash}`));
    return (0, neverthrow_1.ok)(true);
}
exports.verifyCycleRecord = verifyCycleRecord;
//# sourceMappingURL=verify.js.map